from jittor import Module
from ..higher_jit.patch import _MonkeyPatchBase
from ..higher_jit.optim import DifferentiableOptimizer
from typing import Dict, Any, Callable
from ..utils.op_utils import stop_grads

from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class NGD(DynamicalSystem):
    """
    Implements the optimization procedure of the Naive Gradient Descent (NGD) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    lower_loop : int
        The number of iterations for lower-level optimization.
    solver_config : Dict[str, Any]
        A dictionary containing configurations for the solver. Expected keys include:

        - "lower_level_opt" (jittor.optim.Optimizer): The optimizer for the lower-level model.
        - "hyper_op" (List[str]): A list of hyper-gradient operations to apply, such as "PTT" or "FOA".
        - "RGT" (Dict): Configuration for Truncated Gradient Iteration (RGT):
            - "truncate_iter" (int): The number of iterations to truncate the gradient computation.

    References
    ----------
    [1] L. Franceschi, P. Frasconi, S. Salzo, R. Grazzi, and M. Pontil, "Bilevel
        programming for hyperparameter optimization and meta-learning", in ICML, 2018.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        lower_loop: int,
        solver_config: Dict[str, Any],
    ):
        super(NGD, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )
        self.truncate_max_loss_iter = "PTT" in solver_config["hyper_op"]
        self.truncate_iters = solver_config["RGT"]["truncate_iter"]
        self.ll_opt = solver_config["lower_level_opt"]
        self.foa = "FOA" in solver_config["hyper_op"]

    def optimize(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        auxiliary_opt: DifferentiableOptimizer,
        current_iter: int,
        next_operation: str = None,
        **kwargs
    ):
        """
        Execute the lower-level optimization procedure using data, models, and patched optimizers.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization. Typically includes training data, targets, and other information required to compute the lower-level (LL) objective.

        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization. Typically includes validation data, targets, and other information required to compute the upper-level (UL) objective.

        auxiliary_model : _MonkeyPatchBase
            A patched lower-level model wrapped by the `higher` library. Used for differentiable optimization in the lower-level procedure.

        auxiliary_opt : DifferentiableOptimizer
            A patched optimizer for the lower-level model, wrapped by the `higher` library. Enables differentiable optimization.

        current_iter : int
            The current iteration number of the optimization process.

        Returns
        -------
        None
        """

        assert next_operation is None, "NGD does not support next_operation"
        if "gda_loss" in kwargs:
            gda_loss = kwargs["gda_loss"]
            alpha = kwargs["alpha"]
            alpha_decay = kwargs["alpha_decay"]
        else:
            gda_loss = None

        # truncate with RGT operation
        if self.truncate_iters > 0:
            ll_backup = [x.clone().stop_grad() for x in self.ll_model.parameters()]
            for _ in range(self.truncate_iters):
                if gda_loss is not None:
                    ll_feed_dict["alpha"] = alpha
                    loss_f = gda_loss(
                        ll_feed_dict, ul_feed_dict, self.ul_model, auxiliary_model
                    )
                    alpha = alpha * alpha_decay
                else:
                    loss_f = self.ll_objective(
                        ll_feed_dict, self.ul_model, auxiliary_model
                    )
                self.ll_opt.step(loss_f)
            for x, y in zip(self.ll_model.parameters(), auxiliary_model.parameters()):
                y.update(x.clone())
            for x, y in zip(ll_backup, self.ll_model.parameters()):
                y.update(x.clone())

        # truncate with PTT method
        if self.truncate_max_loss_iter:
            ul_loss_list = []
            for _ in range(self.lower_loop):
                if gda_loss is not None:
                    ll_feed_dict["alpha"] = alpha
                    loss_f = gda_loss(
                        ll_feed_dict, ul_feed_dict, self.ul_model, auxiliary_model
                    )
                    alpha = alpha * alpha_decay
                else:
                    loss_f = self.ll_objective(
                        ll_feed_dict, self.ul_model, auxiliary_model
                    )
                auxiliary_opt.step(loss_f)
                upper_loss = self.ul_objective(
                    ul_feed_dict, self.ul_model, auxiliary_model
                )
                ul_loss_list.append(upper_loss.item())
            ll_step_with_max_ul_loss = ul_loss_list.index(max(ul_loss_list))
            return ll_step_with_max_ul_loss + 1

        for _ in range(self.lower_loop - self.truncate_iters):
            if gda_loss is not None:
                ll_feed_dict["alpha"] = alpha
                loss_f = gda_loss(
                    ll_feed_dict, ul_feed_dict, self.ul_model, auxiliary_model
                )
                alpha = alpha * alpha_decay
            else:
                loss_f = self.ll_objective(ll_feed_dict, self.ul_model, auxiliary_model)
            auxiliary_opt.step(loss_f)
            auxiliary_opt.step(loss_f, grad_callback=stop_grads if self.foa else None)
        return -1


# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Utility functions for components of ``higher_jit``.

"""

import typing as _typing
import jittor as jit

_T = _typing.TypeVar("_T")
_U = _typing.TypeVar("_U")


def _copy_tensor(
    t: jit.Var, safe_copy: bool, device: _typing.Optional[str] = None
) -> jit.Var:
    if safe_copy:
        t = t.clone().detach()
        t.stop_grad = not t.is_stop_grad()
    else:
        t = t.detach()
        t.stop_grad = not t.is_stop_grad()
    t = t if device is None else t.to(device)
    return t


def _recursive_copy_and_cast(
    target: _typing.Union[list, tuple, dict, set, jit.Var],
    device: _typing.Optional[str],
) -> jit.Var:
    def map_fn(x):
        if isinstance(
            x, jit.Var
        ):  # Replacing _torch.is_tensor(x) with isinstance(x, jit.Var)
            return _copy_tensor(x, True, device=device)
        else:
            return x

    return _recursive_map(target, map_fn)


def _recursive_map(
    target: _typing.Union[list, tuple, dict, set, _T],
    map_fn: _typing.Callable[[_T], _U],
) -> _typing.Union[list, tuple, dict, set, _U]:
    if isinstance(target, list):
        return type(target)([_recursive_map(x, map_fn) for x in target])
    elif isinstance(target, tuple):
        return type(target)([_recursive_map(x, map_fn) for x in target])
    elif isinstance(target, dict):
        return type(target)({k: _recursive_map(v, map_fn) for k, v in target.items()})
    elif isinstance(target, set):
        return type(target)({_recursive_map(x, map_fn) for x in target})
    else:
        return map_fn(target)


def _is_container(target: _typing.Any) -> bool:
    flag = (
        isinstance(target, list)
        or isinstance(target, tuple)
        or isinstance(target, dict)
        or isinstance(target, set)
    )
    return flag


def _find_param_in_list(
    param: jit.Var, l: _typing.Iterable[jit.Var]
) -> _typing.Optional[int]:
    for i, p in enumerate(l):
        if p is param:
            return i
    else:
        return None


def _get_param_mapping(
    module: jit.Module, seen: _typing.List[jit.Var], mapping: _typing.List[int]
) -> _typing.List[int]:
    # for param in module._parameters.values():
    # k=0
    # print("test_length!!!!!!!!!!!!",len(module.named_parameters(False)))
    for _, param in module.named_parameters(False):
        if param is None:
            continue
        # k+=1
        found = _find_param_in_list(param, seen)
        # print("really",k, found)
        if found is None:
            mapping.append(len(seen))
            seen.append(param)
        else:
            mapping.append(found)

    for name, child in module._modules.items():
        _ = _get_param_mapping(child, seen, mapping)

    return mapping


def flatten(x: _typing.Any) -> _typing.List[_typing.Any]:
    r"""Returns a flattened list of objects from a nested structure."""
    l: _typing.List[_typing.Any] = []
    if isinstance(x, dict):
        for y in x.values():
            l.extend(flatten(y))
    elif isinstance(x, list) or isinstance(x, set) or isinstance(x, tuple):
        for y in x:
            l.extend(flatten(y))
    else:
        l.append(x)
    return l


def get_func_params(
    module: jit.Module, device: _typing.Optional[str] = None, safe_copy: bool = True
) -> _typing.List[jit.Var]:
    r"""Returns a detached copy of module parameters which requires gradient."""
    params = [_copy_tensor(p, safe_copy, device) for p in module.parameters()]
    return params



from typing import List, Dict
from boat_jit.utils import DynamicalSystemRules, ResultStore
from boat_jit.operation_registry import get_registered_operation

importlib = __import__("importlib")


class SequentialDS:
    """
    A dynamically created class for sequential hyper-gradient operations.

    Attributes
    ----------
    gradient_instances : List[object]
        A list of gradient operator instances, each implementing an `optimize` method.
    custom_order : List[str]
        A custom-defined order for executing the gradient operators.
    result_store : ResultStore
        An instance of the `ResultStore` class for storing intermediate and final results.
    """

    def __init__(self, ordered_instances: List[object], custom_order: List[str]):
        """
        Initialize the SequentialDS class with gradient operator instances and a custom execution order.

        Parameters
        ----------
        ordered_instances : List[object]
            A list of gradient operator instances to be executed sequentially.
        custom_order : List[str]
            A list defining the custom execution order of the gradient operators.
        """
        self.gradient_instances = ordered_instances
        self.custom_order = custom_order
        self.result_store = ResultStore()  # Use a dedicated result store

    def optimize(self, **kwargs) -> List[Dict]:
        """
        Compute gradients sequentially using the ordered gradient operator instances.

        Parameters
        ----------
        **kwargs : dict
            Arbitrary keyword arguments required for gradient computations.

        Returns
        -------
        List[Dict]
            A list of dictionaries containing results for each gradient operator.

        Notes
        -----
        - The results of each gradient operator are passed as inputs to the subsequent operator.
        - Results are stored in the `ResultStore` instance for further use or analysis.

        Example
        -------
        >>> gradient_instances = [GradientOp1(), GradientOp2()]
        >>> custom_order = ["op1", "op2"]
        >>> sequential_ds = SequentialDS(gradient_instances, custom_order)
        >>> results = sequential_ds.optimize(input_data=data)
        """
        self.result_store.clear()  # Reset the result store
        intermediate_result = None

        for idx, gradient_instance in enumerate(self.gradient_instances):
            # Compute the gradient, passing the intermediate result as input
            result = gradient_instance.optimize(
                **(kwargs if idx == 0 else intermediate_result),
                next_operation=(
                    self.custom_order[idx + 1]
                    if idx + 1 < len(self.custom_order)
                    else None
                ),
            )
            # Store the result
            self.result_store.add(f"dynamic_results_{idx}", result)
            intermediate_result = result

        return self.result_store.get_results()


def makes_functional_dynamical_system(
    custom_order: List[str], **kwargs
) -> SequentialDS:
    """
    Dynamically create a SequentialHyperGradient object with ordered gradient operators.

    Parameters
    ----------
    custom_order : List[str]
        User-defined operator order.

    Returns
    -------
    SequentialHyperGradient
        An instance with ordered gradient operators and result management.
    """
    # Load the predefined gradient order
    gradient_order = DynamicalSystemRules.get_gradient_order()

    # Adjust custom order based on predefined gradient order
    adjusted_order = validate_and_adjust_order(custom_order, gradient_order)

    # Dynamically load classes
    gradient_classes = {}
    # module = importlib.import_module("boat.dynamic_ol")
    for op in custom_order:
        gradient_classes[op] = get_registered_operation(op)

    # Reorder classes according to adjusted order
    ordered_instances = [gradient_classes[op](**kwargs) for op in adjusted_order]

    # Return the enhanced sequential hyper-gradient class
    return SequentialDS(ordered_instances, custom_order)


def validate_and_adjust_order(
    custom_order: List[str], gradient_order: List[List[str]]
) -> List[str]:
    """
    Validate and adjust the custom order to align with the predefined gradient operator groups.

    Parameters
    ----------
    custom_order : List[str]
        The user-defined order of gradient operators.
    gradient_order : List[List[str]]
        The predefined grouping of gradient operators, specifying valid order constraints.

    Returns
    -------
    List[str]
        A validated and adjusted list of gradient operators that conforms to the predefined order.

    Notes
    -----
    - The function filters out invalid operators from `custom_order` that do not exist in `gradient_order`.
    - It ensures that the returned order follows the precedence rules defined in `gradient_order`.

    Example
    -------
    >>> custom_order = ["op1", "op3", "op2"]
    >>> gradient_order = [["op1", "op2"], ["op3"]]
    >>> adjusted_order = validate_and_adjust_order(custom_order, gradient_order)
    >>> print(adjusted_order)
    ['op1', 'op2', 'op3']
    """
    # Create a set of valid operators for quick lookup
    valid_operators = {op for group in gradient_order for op in group}

    # Filter out invalid operators
    custom_order = [op for op in custom_order if op in valid_operators]

    # Adjust order to follow gradient_order
    adjusted_order = []
    for group in gradient_order:
        for op in group:
            if op in custom_order:
                adjusted_order.append(op)

    return adjusted_order



from .hyper_gradient import HyperGradient
from .fd import FD
from .cg import CG
from .ns import NS
from .iga import IGA
from .rad import RAD
from .iad import IAD
from .foa import FOA
from .rgt import RGT
from .ptt import PTT
from .sequential_hg import makes_functional_hyper_operation



from jittor import Module
from typing import Callable
from ..higher_jit.patch import _MonkeyPatchBase
from ..higher_jit.optim import DifferentiableOptimizer
from typing import Dict, Any, Callable

from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class GDA(DynamicalSystem):
    """
    Implements the optimization procedure of the Gradient Descent Aggregation (GDA) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    lower_loop : int
        The number of iterations for lower-level optimization.
    solver_config : Dict[str, Any]
        A dictionary containing configurations for the solver. Expected keys include:

        - "GDA" (Dict): Configuration for the GDA algorithm:
            - "alpha_init" (float): Initial learning rate for the GDA updates.
            - "alpha_decay" (float): Decay rate for the learning rate.
        - "gda_loss" (Callable): The loss function used in the GDA optimization.

    References
    ----------
    [1] R. Liu, P. Mu, X. Yuan, S. Zeng, and J. Zhang, "A generic first-order algorithmic framework for bi-level programming beyond lower-level singleton", in ICML, 2020.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        lower_loop: int,
        solver_config: Dict[str, Any],
    ):
        super(GDA, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )
        self.alpha = solver_config["GDA"]["alpha_init"]
        self.alpha_decay = solver_config["GDA"]["alpha_decay"]
        self.gda_loss = solver_config["gda_loss"]

    def optimize(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        auxiliary_opt: DifferentiableOptimizer,
        current_iter: int,
        next_operation: str = None,
        **kwargs
    ):
        """
        Execute the lower-level optimization procedure using provided data, models, and patched optimizers.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            Typically includes training data, targets, and other information required to compute the LL objective.
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            Typically includes validation data, targets, and other information required to compute the UL objective.
        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library.
            This model is used for differentiable optimization in the lower-level procedure.
        auxiliary_opt : DifferentiableOptimizer
            A patched optimizer for the lower-level model, wrapped by the `higher` library.
            Allows differentiable optimization.
        current_iter : int
            The current iteration number of the optimization process.
        next_operation : str, optional
            Specifies the next operation to be executed in the optimization pipeline.
            Default is None.
        **kwargs : dict
            Additional parameters required for the optimization procedure.

        Returns
        -------
        dict
            A dictionary containing:
                - "ll_feed_dict" : Dict
                    Lower-level feed dictionary.
                - "ul_feed_dict" : Dict
                    Upper-level feed dictionary.
                - "auxiliary_model" : _MonkeyPatchBase
                    Patched lower-level model.
                - "auxiliary_opt" : DifferentiableOptimizer
                    Patched lower-level optimizer.
                - "current_iter" : int
                    Current iteration number.
                - "gda_loss" : callable
                    Gradient Descent Aggregation (GDA) loss function.
                - "alpha" : float
                    Coefficient used in the GDA operation, typically in (0, 1).
                - "alpha_decay" : float
                    Decay factor for the coefficient `alpha`.

        Raises
        ------
        AssertionError
            If `next_operation` is not defined.
            If `alpha` is not in the range (0, 1).
            If `gda_loss` is not properly defined.

        Notes
        -----
        - The method assumes that `gda_loss` is defined and accessible from the instance attributes.
        - The coefficient `alpha` and its decay rate `alpha_decay` must be properly configured.
        """

        assert next_operation is not None, "Next operation should be defined."
        assert (self.alpha > 0) and (
            self.alpha < 1
        ), "Set the coefficient alpha properly in (0,1)."
        assert (
            self.gda_loss is not None
        ), "Define the gda_loss properly in loss_func.py."
        return {
            "ll_feed_dict": ll_feed_dict,
            "ul_feed_dict": ul_feed_dict,
            "auxiliary_model": auxiliary_model,
            "auxiliary_opt": auxiliary_opt,
            "current_iter": current_iter,
            "gda_loss": self.gda_loss,
            "alpha": self.alpha,
            "alpha_decay": self.alpha_decay,
            **kwargs,
        }



Operation_REGISTRY = {}


def register_class(cls):
    """
    Register a new operation class to the global registry.
    """
    Operation_REGISTRY[cls.__name__] = cls
    return cls


def get_registered_operation(name):
    """
    Retrieve a registered operation class by name.
    """
    if name not in Operation_REGISTRY:
        raise ValueError(f"Class '{name}' is not registered.")
    return Operation_REGISTRY[name]



from abc import abstractmethod
from boat_jit.operation_registry import register_class

importlib = __import__("importlib")


@register_class
class HyperGradient(object):
    """
    Base class for computing hyper-gradients of upper-level variables in bilevel optimization problems.

    This class provides an abstract interface for hyper-gradient computation that can be extended
    for specific methods such as Conjugate Gradient, Finite Differentiation, or First-Order Approximation.

    Parameters
    ----------
    ll_objective : callable
        The lower-level objective function of the bilevel optimization problem.

    ul_objective : callable
        The upper-level objective function of the bilevel optimization problem.

    ul_model : jittor.Module
        The upper-level model of the bilevel optimization problem.

    ll_model : jittor.Module
        The lower-level model of the bilevel optimization problem.

    ll_var : List[jittor.Var]
        A list of variables optimized with the lower-level objective.

    ul_var : List[jittor.Var]
        A list of variables optimized with the upper-level objective.

    solver_config : dict
        Dictionary containing configurations for the solver.
    """

    def __init__(
        self,
        ll_objective,
        ul_objective,
        ul_model,
        ll_model,
        ll_var,
        ul_var,
        solver_config,
    ):
        self.ll_objective = ll_objective
        self.ul_objective = ul_objective
        self.ul_model = ul_model
        self.ll_model = ll_model
        self.ll_var = ll_var
        self.ul_var = ul_var
        self.solver_configs = solver_config

    @abstractmethod
    def compute_gradients(self, **kwargs):
        pass



import time
import copy
from typing import Dict, Any, Callable
from boat_jit.utils.op_utils import (
    copy_parameter_from_list,
    average_grad,
    manual_update,
)

import jittor as jit
from jittor.optim import Optimizer
import boat_jit.higher_jit as higher


importlib = __import__("importlib")
from boat_jit.operation_registry import get_registered_operation
from boat_jit.dynamic_ol import makes_functional_dynamical_system
from boat_jit.hyper_ol import makes_functional_hyper_operation


def _load_loss_function(loss_config: Dict[str, Any]) -> Callable:
    """
    Dynamically load a loss function from the provided configuration.

    :param loss_config: Dictionary with keys:
        - "function": Path to the loss function (e.g., "module.path.to_function").
        - "params": Parameters to be passed to the loss function.
    :type loss_config: Dict[str, Any]

    :returns: Loaded loss function ready for use.
    :rtype: Callable
    """

    module_name, func_name = loss_config["function"].rsplit(".", 1)
    module = importlib.import_module(module_name)
    func = getattr(module, func_name)

    # Return a wrapper function that can accept both positional and keyword arguments
    return lambda *args, **kwargs: func(
        *args, **{**loss_config.get("params", {}), **kwargs}
    )


class Problem:
    """
    Enhanced bi-level optimization problem class supporting flexible loss functions and operation configurations.
    """

    def __init__(self, config: Dict[str, Any], loss_config: Dict[str, Any]):
        """
        Initialize the Problem instance.

        :param config: Configuration dictionary for the optimization setup.
            - "fo_gm": First Order Gradient based Method (optional), e.g., ["VSM"], ["VFM"], ["MESM"].
            - "dynamic_op": List of dynamic operations (optional), e.g., ["NGD"], ["NGD", "GDA"], ["NGD", "GDA", "DI"].
            - "hyper_op": Hyper-optimization method (optional), e.g., ["RAD"], ["RAD", "PTT"], ["IAD", "NS", "PTT"].
            - "lower_level_loss": Configuration for the lower-level loss function based on the json file configuration.
            - "upper_level_loss": Configuration for the upper-level loss function based on the json file configuration.
            - "lower_level_model": The lower-level model to be optimized.
            - "upper_level_model": The upper-level model to be optimized.
            - "lower_level_var": Variables in the lower-level model.
            - "upper_level_var": Variables in the upper-level model.
            - "device": Device configuration (e.g., "cpu", "cuda").
        :type config: Dict[str, Any]

        :param loss_config: Loss function configuration dictionary.
            - "lower_level_loss": Configuration for the lower-level loss function.
            - "upper_level_loss": Configuration for the upper-level loss function.
            - "GDA_loss": Configuration for GDA loss function (optional).
        :type loss_config: Dict[str, Any]

        :returns: None
        """
        self._fo_gm = config["fo_gm"]
        self._dynamic_op = config["dynamic_op"]
        self._hyper_op = config["hyper_op"]
        self._ll_model = config["lower_level_model"]
        self._ul_model = config["upper_level_model"]
        self._ll_var = config["lower_level_var"]
        self._ul_var = config["upper_level_var"]
        self.boat_configs = config
        if config["dynamic_op"] is not None:
            if "GDA" in config["dynamic_op"]:
                assert (
                    loss_config.get("gda_loss", None) is not None
                ), "Set the 'gda_loss' in loss_config properly."
                self.boat_configs["gda_loss"] = _load_loss_function(
                    loss_config["gda_loss"]
                )
        self._lower_loop = config.get("lower_iters", 10)
        self._lower_opt = self.boat_configs["lower_level_opt"]
        self._upper_opt = self.boat_configs["upper_level_opt"]
        self._ll_loss = _load_loss_function(loss_config["lower_level_loss"])
        self._ul_loss = _load_loss_function(loss_config["upper_level_loss"])
        self._ll_solver = None
        self._ul_solver = None
        self._lower_init_opt = None
        self._fo_gm_solver = None
        self._log_results = []
        self._track_opt_traj = False

    def build_ll_solver(self):
        """
        Configure the lower-level solver.

        :returns: None
        """
        if self.boat_configs["fo_gm"] is None:
            assert (self.boat_configs["dynamic_op"] is not None) and (
                self.boat_configs["hyper_op"] is not None
            ), "Set 'dynamic_op' and 'hyper_op' properly."

            self.check_status()
            if "DM" in self._dynamic_op:
                self.boat_configs["DM"]["auxiliary_v"] = [
                    jit.zeros_like(param) for param in self._ll_var
                ]
                self.boat_configs["DM"]["auxiliary_v_opt"] = jit.nn.SGD(
                    self.boat_configs["DM"]["auxiliary_v"],
                    lr=self.boat_configs["DM"]["auxiliary_v_lr"],
                )
            sorted_ops = sorted([op.upper() for op in self._dynamic_op])
            self._ll_solver = makes_functional_dynamical_system(
                custom_order=sorted_ops,
                ll_objective=self._ll_loss,
                ul_objective=self._ul_loss,
                ll_model=self._ll_model,
                ul_model=self._ul_model,
                lower_loop=self._lower_loop,
                solver_config=self.boat_configs,
            )
            if "DI" in self.boat_configs["dynamic_op"]:
                self._lower_init_opt = copy.deepcopy(self._lower_opt)
                for _ in range(len(self._lower_init_opt.param_groups)):
                    self._lower_init_opt.param_groups[_][
                        "params"
                    ] = self._lower_opt.param_groups[_]["params"]
                    self._lower_init_opt.param_groups[_]["lr"] = self.boat_configs[
                        "DI"
                    ]["lr"]
        else:
            self._fo_gm_solver = get_registered_operation(
                "%s" % self.boat_configs["fo_gm"]
            )(
                ll_objective=self._ll_loss,
                ul_objective=self._ul_loss,
                ll_model=self._ll_model,
                ul_model=self._ul_model,
                lower_loop=self._lower_loop,
                ll_var=self._ll_var,
                ul_var=self._ul_var,
                solver_config=self.boat_configs,
            )
        return self

    def build_ul_solver(self):
        """
        Configure the lower-level solver.

        :returns: None
        """
        if self.boat_configs["fo_gm"] is None:
            assert (
                self.boat_configs["hyper_op"] is not None
            ), "Choose FOGM based methods from ['VSM'],['VFM'],['MESM'] or set 'dynamic_ol' and 'hyper_ol' properly."
            sorted_ops = sorted([op.upper() for op in self._hyper_op])
            self._ul_solver = makes_functional_hyper_operation(
                custom_order=sorted_ops,
                ul_objective=self._ul_loss,
                ll_objective=self._ll_loss,
                ll_model=self._ll_model,
                ul_model=self._ul_model,
                ll_var=self._ll_var,
                ul_var=self._ul_var,
                solver_config=self.boat_configs,
            )
        else:
            assert (
                self.boat_configs["fo_gm"] is not None
            ), "Choose FOGM based methods from ['VSM','VFM','MESM'] or set 'dynamic_ol' and 'hyper_ol' properly."

        return self

    def run_iter(
        self,
        ll_feed_dict: Dict[str, jit.Var],
        ul_feed_dict: Dict[str, jit.Var],
        current_iter: int,
    ) -> tuple:
        """
        Run a single iteration of the bi-level optimization process.

        :param ll_feed_dict: Dictionary containing the real-time data and parameters
            fed for the construction of the lower-level (LL) objective.

            Example::

                {
                    "image": train_images,
                    "text": train_texts,
                    "target": train_labels  # Optional
                }

        :type ll_feed_dict: Dict[str, Tensor]

        :param ul_feed_dict: Dictionary containing the real-time data and parameters
            fed for the construction of the upper-level (UL) objective.

            Example::

                {
                    "image": val_images,
                    "text": val_texts,
                    "target": val_labels  # Optional
                }

        :type ul_feed_dict: Dict[str, Tensor]

        :param current_iter: The current iteration number.
        :type current_iter: int

        :notes:
            - When `accumulate_grad` is set to True, you need to pack the data of
              each batch based on the format above.
            - In that case, pass `ll_feed_dict` and `ul_feed_dict` as lists of
              dictionaries, i.e., `[Dict[str, Tensor]]`.

        :returns: A tuple containing:
            - **loss** (*float*): The loss value for the current iteration.
            - **run_time** (*float*): The total time taken for the iteration.

        :rtype: tuple
        """
        if self.boat_configs["fo_gm"] is not None:
            start_time = time.perf_counter()
            self._log_results.append(
                self._fo_gm_solver.optimize(ll_feed_dict, ul_feed_dict, current_iter)
            )
            run_time = time.perf_counter() - start_time
        else:
            run_time = 0
            if self.boat_configs["accumulate_grad"]:
                for batch_ll_feed_dict, batch_ul_feed_dict in zip(
                    ll_feed_dict, ul_feed_dict
                ):
                    with higher.innerloop_ctx(
                        self._ll_model, self._lower_opt, copy_initial_weights=False, track_higher_grads=self._track_opt_traj
                    ) as (auxiliary_model, auxiliary_opt):
                        forward_time = time.perf_counter()
                        dynamic_results = self._ll_solver.optimize(
                            ll_feed_dict=batch_ll_feed_dict,
                            ul_feed_dict=batch_ul_feed_dict,
                            auxiliary_model=auxiliary_model,
                            auxiliary_opt=auxiliary_opt,
                            current_iter=current_iter,
                        )
                        self._log_results.append(dynamic_results)
                        max_loss_iter = list(dynamic_results[-1].values())[-1]
                        forward_time = time.perf_counter() - forward_time
                        backward_time = time.perf_counter()
                        self._log_results.append(
                            self._ul_solver.compute_gradients(
                                ll_feed_dict=batch_ll_feed_dict,
                                ul_feed_dict=batch_ll_feed_dict,
                                auxiliary_model=auxiliary_model,
                                max_loss_iter=max_loss_iter,
                            )
                        )
                        backward_time = time.perf_counter() - backward_time
                    run_time += forward_time + backward_time
                average_grad(self._ul_model, len(ll_feed_dict))
            else:
                with higher.innerloop_ctx(
                    self._ll_model, self._lower_opt, copy_initial_weights=False, track_higher_grads=self._track_opt_traj
                ) as (auxiliary_model, auxiliary_opt):
                    forward_time = time.perf_counter()
                    dynamic_results = self._ll_solver.optimize(
                        ll_feed_dict=ll_feed_dict,
                        ul_feed_dict=ul_feed_dict,
                        auxiliary_model=auxiliary_model,
                        auxiliary_opt=auxiliary_opt,
                        current_iter=current_iter,
                    )
                    self._log_results.append(dynamic_results)
                    max_loss_iter = list(dynamic_results[-1].values())[-1]
                    forward_time = time.perf_counter() - forward_time
                    print('forward_time',forward_time)
                    backward_time = time.perf_counter()
                    if "DM" not in self._dynamic_op:
                        self._log_results.append(
                            self._ul_solver.compute_gradients(
                                ll_feed_dict=ll_feed_dict,
                                ul_feed_dict=ul_feed_dict,
                                auxiliary_model=auxiliary_model,
                                max_loss_iter=max_loss_iter,
                            )
                        )
                    else:
                        self._log_results.append(
                            self._ul_loss(ul_feed_dict, self._ul_model, auxiliary_model)
                        )
                    backward_time = time.perf_counter() - backward_time
                    print('backward_time', backward_time)
                    if self.boat_configs["copy_last_param"]:
                        copy_parameter_from_list(
                            self._ll_model,
                            list(auxiliary_model.parameters(time=-1)),
                        )
                # update the dynamic initialization of lower-level variables
                if "DI" in self.boat_configs["dynamic_op"]:
                    manual_update(
                        self._lower_init_opt, self._lower_opt.param_groups[0]["params"]
                    )
                run_time = forward_time + backward_time
        if not self.boat_configs["return_grad"]:
            manual_update(self._upper_opt, self._ul_var)
        else:
            return [var._custom_grad for var in list(self._ul_var)], run_time

        return self._log_results, run_time

    def set_track_trajectory(self, track_traj=True):
        self._track_opt_traj = track_traj

    def check_status(self):
        if any(item in self._hyper_op for item in ["PTT", "IAD", "RAD"]):
            self.set_track_trajectory(True)
        if "DM" in self.boat_configs["dynamic_op"]:
            assert (self.boat_configs["hyper_op"] == ["RAD"]) or (
                self.boat_configs["hyper_op"] == ["CG"]
            ), "When 'DM' is chosen, set the 'truncate_iter' properly."
        if "RGT" in self.boat_configs["hyper_op"]:
            assert (
                self.boat_configs["RGT"]["truncate_iter"] > 0
            ), "When 'RGT' is chosen, set the 'truncate_iter' properly ."
        if self.boat_configs["accumulate_grad"]:
            assert (
                "IAD" in self.boat_configs["hyper_op"]
            ), "When using 'accumulate_grad', only 'IAD' based methods are supported."
        if self.boat_configs["GDA"]["alpha_init"] > 0.0:
            assert (
                0.0 < self.boat_configs["GDA"]["alpha_decay"] <= 1.0
            ), "Parameter 'alpha_decay' used in method BDA should be in the interval (0,1)."
        if "FD" in self._hyper_op:
            assert (
                self.boat_configs["RGT"]["truncate_iter"] == 0
            ), "One-stage method doesn't need trajectory truncation."

        def check_model_structure(base_model, meta_model):
            for param1, param2 in zip(base_model.parameters(), meta_model.parameters()):
                if (
                    (param1.shape != param2.shape)
                    or (param1.dtype != param2.dtype)
                    or (param1.device != param2.device)
                ):
                    return False
            return True

        if "IAD" in self._hyper_op:
            assert check_model_structure(self._ll_model, self._ul_model), (
                "With IAD or FOA operation, 'upper_level_model' and 'lower_level_model' have the same structure, "
                "and 'lower_level_var' and 'upper_level_var' are the same group of variables."
            )
        assert (("DI" in self._dynamic_op) ^ ("IAD" in self._hyper_op)) or (
            ("DI" not in self._dynamic_op) and ("IAD" not in self._hyper_op)
        ), "Only one of the 'PTT' and 'RGT' methods could be chosen."
        assert (
            0.0 <= self.boat_configs["GDA"]["alpha_init"] <= 1.0
        ), "Parameter 'alpha' used in method BDA should be in the interval (0,1)."
        assert (
            self.boat_configs["RGT"]["truncate_iter"] < self.boat_configs["lower_iters"]
        ), "The value of 'truncate_iter' shouldn't be greater than 'lower_loop'."



from setuptools import setup, find_packages
from os import path

current_directory = path.abspath(path.dirname(__file__))

with open("README.md",  "r",encoding="utf-8",) as fh:
    long_description = fh.read()


def get_install_requirements():
    requirements_path = path.join(current_directory, "requirements_jit.txt")
    with open(requirements_path, encoding="utf-8") as fp:
        return fp.read().splitlines()


setup(
    name="boat-jit",
    version="1.0.2",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/callous-youth/BOAT/tree/boat_jit",
    license="MIT",
    keywords=[
        "Bilevel-optimization",
        "Learning and vision",
        "Python",
        "Deep learning",
        "Jittor",
    ],
    install_requires=[
        "setuptools",
        "numpy",
        "tqdm",
        "scikit-learn>=1.3",
        "jittor",
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8.0",
    author="Yaohua Liu, Xianghao Jiao, Risheng Liu",
    author_email="liuyaohua.918@gmail.com",
    description="A Bilevel Optimization Toolkit in Python for Learning and Vision Tasks Based on Jittor",
)



from .dynamical_system import DynamicalSystem
import jittor as jit
from jittor import Module
from ..higher_jit.patch import _MonkeyPatchBase
from ..higher_jit.optim import DifferentiableOptimizer
from typing import Dict, Any, Callable
from ..utils.op_utils import (
    update_tensor_grads,
    grad_unused_zero,
    list_tensor_norm,
    list_tensor_matmul,
    custom_grad,
    manual_update,
)

from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class DM(DynamicalSystem):
    """
    Implements the lower-level optimization procedure for Dual Multiplier (DM) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    lower_loop : int
        The number of iterations for the lower-level optimization process.
    solver_config : Dict[str, Any]
        A dictionary containing configurations for the optimization solver, including
        hyperparameters and specific settings for NGD, GDA, and DM.

    References
    ----------
    [1] Liu R, Liu Y, Yao W, et al., "Averaged method of multipliers for bi-level optimization without lower-level
        strong convexity," ICML, 2023.
    """

    def __init__(
        self,
        ll_objective: Callable,
        lower_loop: int,
        ul_model: Module,
        ul_objective: Callable,
        ll_model: Module,
        solver_config: Dict[str, Any],
    ):
        super(DM, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )
        self.truncate_max_loss_iter = "PTT" in solver_config["hyper_op"]
        self.alpha = solver_config["GDA"]["alpha_init"]
        self.alpha_decay = solver_config["GDA"]["alpha_decay"]
        self.truncate_iters = solver_config["RGT"]["truncate_iter"]
        self.ll_opt = solver_config["lower_level_opt"]
        self.ul_opt = solver_config["upper_level_opt"]
        self.auxiliary_v = solver_config["DM"]["auxiliary_v"]
        self.auxiliary_v_opt = solver_config["DM"]["auxiliary_v_opt"]
        self.auxiliary_v_lr = solver_config["DM"]["auxiliary_v_lr"]
        self.tau = solver_config["DM"]["tau"]
        self.p = solver_config["DM"]["p"]
        self.mu0 = solver_config["DM"]["mu0"]
        self.eta = solver_config["DM"]["eta0"]
        self.strategy = solver_config["DM"]["strategy"]
        self.hyper_op = solver_config["hyper_op"]
        self.gda_loss = solver_config["gda_loss"]

    def optimize(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        auxiliary_opt: DifferentiableOptimizer,
        current_iter: int,
        next_operation: str = None,
        **kwargs
    ):
        """
        Executes the lower-level optimization procedure with support for NGD, GDA, and RAD operations.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization. Typically includes:
                - "data" : Training input data.
                - "target" : Training target data (optional, depending on the task).
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization. Typically includes:
                - "data" : Validation input data.
                - "target" : Validation target data (optional, depending on the task).
        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library. Used for differentiable optimization.
        auxiliary_opt : DifferentiableOptimizer
            A patched optimizer for the lower-level model, wrapped by the `higher` library. Enables differentiable optimization steps.
        current_iter : int
            The current iteration number in the optimization process.
        next_operation : str, optional
            Specifies the next operation in the optimization process. Must be `None` for NGD. (default: None)
        kwargs : dict
            Additional keyword arguments for the optimization process.

        Returns
        -------
        int
            Returns `-1` upon successful completion of the optimization process.

        Notes
        -----
        - For GDA operations, this method supports three strategies: 's1', 's2', and 's3'.
        - When using RAD in `hyper_op`, a higher-order gradient adjustment is applied to the auxiliary variables.
        - Ensure that `next_operation` is `None` for NGD, as it does not support additional operations.

        Raises
        ------
        AssertionError
            If `next_operation` is not `None` for NGD or if an unsupported strategy is specified for GDA.
        """
        assert next_operation is None, "NGD does not support next_operation"
        if "gda_loss" in kwargs:
            gda_loss = kwargs["gda_loss"]
            assert self.strategy in [
                "s1",
                "s2",
                "s3",
            ], "Three strategies are supported for DM operation, including ['s1','s2','s3']."
            if self.strategy == "s1":
                self.alpha = self.mu0 * 1 / (current_iter + 1) ** (1 / self.p)
                self.eta = (
                    (current_iter + 1) ** (-0.5 * self.tau)
                    * self.alpha**2
                    * self.ll_opt.defaults["lr"]
                )
                x_lr = (
                    (current_iter + 1) ** (-1.5 * self.tau)
                    * self.alpha**7
                    * self.ll_opt.defaults["lr"]
                )
            elif self.strategy == "s2":
                self.alpha = self.mu0 * 1 / (current_iter + 1) ** (1 / self.p)
                self.eta = (
                    (current_iter + 1) ** (-0.5 * self.tau)
                    * self.alpha
                    * self.ll_opt.defaults["lr"]
                )
                x_lr = (
                    (current_iter + 1) ** (-1.5 * self.tau)
                    * self.alpha**5
                    * self.ll_opt.defaults["lr"]
                )
            elif self.strategy == "s3":
                self.alpha = self.mu0 * 1 / (current_iter + 1) ** (1 / self.p)
                self.eta = (current_iter + 1) ** (
                    -0.5 * self.tau
                ) * self.ll_opt.defaults["lr"]
                x_lr = (
                    (current_iter + 1) ** (-1.5 * self.tau)
                    * self.alpha**3
                    * self.ll_opt.defaults["lr"]
                )
            for params in self.ul_opt.param_groups:
                params["lr"] = x_lr
        else:
            gda_loss = None
            assert (
                self.strategy == "s1"
            ), "Only 's1' strategy is supported for DM without GDA operation."

            x_lr = (
                self.ul_opt.defaults["lr"]
                * (current_iter + 1) ** (-self.tau)
                * self.ll_opt.defaults["lr"]
            )
            eta = (
                self.eta
                * (current_iter + 1) ** (-0.5 * self.tau)
                * self.ll_opt.defaults["lr"]
            )
            for params in self.auxiliary_v_opt.param_groups:
                params["lr"] = eta
            for params in self.ul_opt.param_groups:
                params["lr"] = x_lr
        #############
        # self.ll_opt.zero_grad()
        # self.auxiliary_v_opt.zero_grad()
        # loss_f = self.ll_objective(ll_feed_dict, self.ul_model, auxiliary_model)
        # upper_loss = self.ul_objective(ul_feed_dict, self.ul_model, auxiliary_model)
        # assert (self.alpha>0) and (self.alpha<1), \
        #         "Set the coefficient alpha properly in (0,1)."
        # loss_full = (1.0 - self.alpha) * loss_f + self.alpha * upper_loss

        if gda_loss is not None:
            ll_feed_dict["alpha"] = self.alpha
            loss_full = self.gda_loss(
                ll_feed_dict, ul_feed_dict, self.ul_model, auxiliary_model
            )
        else:
            loss_full = self.ll_objective(ll_feed_dict, self.ul_model, auxiliary_model)

        grad_y_temp = jit.grad(
            loss_full, list(auxiliary_model.parameters()), retain_graph=True
        )
        upper_loss = self.ul_objective(ul_feed_dict, self.ul_model, auxiliary_model)
        grad_outer_params = grad_unused_zero(
            upper_loss, list(auxiliary_model.parameters()), retain_graph=True
        )
        grads_phi_params = grad_unused_zero(
            loss_full, list(auxiliary_model.parameters()), retain_graph=True
        )
        grads = custom_grad(
            grads_phi_params,
            list(self.ul_model.parameters()),
            self.auxiliary_v,
            retain_graph=True,
        )  # dx (dy f) v
        grad_outer_hparams = grad_unused_zero(
            upper_loss, list(self.ul_model.parameters())
        )

        if "RAD" in self.hyper_op:
            vsp = custom_grad(
                grads_phi_params,
                list(auxiliary_model.parameters()),
                grad_outputs=self.auxiliary_v,
            )  # dy (dy f) v=d2y f v

            for v0, v, gow in zip(self.auxiliary_v, vsp, grad_outer_params):
                v0._custom_grad = v - gow
            update_tensor_grads(list(self.ll_model.parameters()), grad_y_temp)
            # self.ll_opt.step()
            manual_update(self.ll_opt, list(self.ll_model.parameters()))
            # self.auxiliary_v_opt.step()
            manual_update(self.auxiliary_v_opt, self.auxiliary_v)

            grads = [
                -g + v if g is not None else v
                for g, v in zip(grads, grad_outer_hparams)
            ]
            update_tensor_grads(list(self.ul_model.parameters()), grads)

        else:
            vsp = custom_grad(
                grads_phi_params,
                list(auxiliary_model.parameters()),
                grad_outputs=self.auxiliary_v,
            )

            tem = [v - gow for v, gow in zip(vsp, grad_outer_params)]

            ita_u = list_tensor_norm(tem) ** 2
            grad_tem = custom_grad(
                grads_phi_params, list(auxiliary_model.parameters()), grad_outputs=tem
            )
            ita_l = list_tensor_matmul(tem, grad_tem)

            ita = ita_u / (ita_l + 1e-12)

            self.auxiliary_v = [
                v0 - ita * v + ita * gow
                for v0, v, gow in zip(self.auxiliary_v, vsp, grad_outer_params)
            ]

            vsp = custom_grad(
                grads_phi_params,
                list(auxiliary_model.parameters()),
                grad_outputs=self.auxiliary_v,
            )

            for v0, v, gow in zip(self.auxiliary_v, vsp, grad_outer_params):
                v0._custom_grad = v - gow

            update_tensor_grads(list(self.ll_model.parameters()), grad_y_temp)
            manual_update(self.ll_opt, list(self.ll_model.parameters()))

            grads = [
                -g + v if g is not None else v
                for g, v in zip(grads, grad_outer_hparams)
            ]
            update_tensor_grads(list(self.ul_model.parameters()), grads)

        return -1



import jittor as jit
from jittor import Module
from typing import List, Callable, Dict
from ..higher_jit.patch import _MonkeyPatchBase
from boat_jit.utils.op_utils import update_tensor_grads

from boat_jit.operation_registry import register_class
from boat_jit.hyper_ol.hyper_gradient import HyperGradient


@register_class
class IGA(HyperGradient):
    """
    Computes the hyper-gradient of the upper-level variables using Implicit Gradient Approximation (IGA) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        List of variables optimized with the lower-level objective.
    ul_var : List[jittor.Var]
        List of variables optimized with the upper-level objective.
    solver_config : Dict[str, Any]
        Dictionary containing solver configurations, including:

        - `alpha_init` (float): Initial learning rate for GDA.
        - `alpha_decay` (float): Decay factor for the GDA learning rate.
        - Optional `gda_loss` (Callable): Custom loss function for GDA, if applicable.
        - `dynamic_op` (List[str]): Specifies dynamic operations, e.g., "DI" for dynamic initialization.

    Attributes
    ----------
    alpha : float
        Initial learning rate for GDA operations, if applicable.
    alpha_decay : float
        Decay factor applied to the GDA learning rate.
    gda_loss : Callable, optional
        Custom loss function for GDA operations, if specified in `solver_config`.
    dynamic_initialization : bool
        Indicates whether dynamic initialization is enabled, based on `dynamic_op`.

    References
    ----------
    [1] Liu R, Gao J, Liu X, et al., "Learning with constraint learning: New perspective, solution strategy and
        various applications," IEEE Transactions on Pattern Analysis and Machine Intelligence, 2024.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict,
    ):
        super(IGA, self).__init__(
            ll_objective,
            ul_objective,
            ul_model,
            ll_model,
            ll_var,
            ul_var,
            solver_config,
        )
        self.alpha = solver_config["GDA"]["alpha_init"]
        self.alpha_decay = solver_config["GDA"]["alpha_decay"]
        self.gda_loss = solver_config.get("gda_loss", None)

        self.dynamic_initialization = "DI" in solver_config["dynamic_op"]

    def compute_gradients(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        max_loss_iter: int = 0,
        hyper_gradient_finished: bool = False,
        next_operation: str = None,
        **kwargs
    ):
        """
        Compute the hyper-gradients of the upper-level variables using the given feed dictionaries and patched models.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization, including training data,
            targets, and other information required for the LL objective computation.
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization, including validation data,
            targets, and other information required for the UL objective computation.
        auxiliary_model : _MonkeyPatchBase
            A patched lower-level model wrapped by the `higher` library, enabling differentiable optimization.
        max_loss_iter : int, optional
            The number of iterations used for backpropagation, by default 0.
        hyper_gradient_finished : bool, optional
            A flag indicating whether the hypergradient computation is finished, by default False.
        next_operation : str, optional
            The next operator for hypergradient calculation. Not supported in this implementation, by default None.
        **kwargs : dict
            Additional arguments, such as:

            - `lower_model_params` : List[torch.nn.Parameter]
                List of parameters for the lower-level model.

        Returns
        -------
        Dict
            A dictionary containing:

            - `upper_loss` : torch.Tensor
                The upper-level objective value after optimization.
            - `hyper_gradient_finished` : bool
                Indicates whether the hypergradient computation is complete.

        Notes
        -----
        - This implementation calculates the Gauss-Newton (GN) loss to refine the gradients using second-order approximations.
        - If `dynamic_initialization` is enabled, the gradients of the lower-level variables are updated with time-dependent parameters.
        - Updates are performed on both lower-level and upper-level variables using computed gradients.

        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """
        assert next_operation is None, "FD does not support next_operation"
        lower_model_params = kwargs.get(
            "lower_model_params", list(auxiliary_model.parameters())
        )
        if self.gda_loss is not None:
            ll_feed_dict["alpha"] = self.alpha * self.alpha_decay**max_loss_iter
            lower_loss = self.gda_loss(
                ll_feed_dict,
                ul_feed_dict,
                self.ul_model,
                auxiliary_model,
                params=lower_model_params,
            )
        else:
            lower_loss = self.ll_objective(
                ll_feed_dict, self.ul_model, auxiliary_model, params=lower_model_params
            )

        upper_loss = self.ul_objective(
            ul_feed_dict, self.ul_model, auxiliary_model, params=lower_model_params
        )
        dFy = jit.grad(upper_loss, lower_model_params, retain_graph=True)
        
        dfy = jit.grad(lower_loss, lower_model_params, retain_graph=True)


        # calculate GN loss
        gFyfy = 0
        gfyfy = 0
        for Fy, fy in zip(dFy, dfy):
            gFyfy = gFyfy + (Fy * fy).sum()
            gfyfy = gfyfy + (fy * fy).sum()
        GN_loss = -gFyfy.detach() / gfyfy.detach() * lower_loss

        if self.dynamic_initialization:
            grads_lower = jit.grad(
                upper_loss, list(auxiliary_model.parameters(time=0)), retain_graph=True
            )
            update_tensor_grads(self.ll_var, grads_lower)

        grads_upper = jit.grad(GN_loss + upper_loss, list(self.ul_var))

        update_tensor_grads(self.ul_var, grads_upper)

        return {"upper_loss": upper_loss.item(), "hyper_gradient_finished": True}



from typing import List, Dict
from boat_jit.utils import HyperGradientRules, ResultStore
from boat_jit.operation_registry import get_registered_operation


class SequentialHG:
    """
    A class for managing sequential hyper-gradient operations.

    This class dynamically organizes and executes a sequence of hyper-gradient computations
    using user-defined and validated orders of gradient operators.

    Parameters
    ----------
    ordered_instances : List[object]
        A list of instantiated gradient operator objects, ordered as per the adjusted sequence.

    custom_order : List[str]
        The user-defined order of gradient operators.
    """

    def __init__(self, ordered_instances: List[object], custom_order: List[str]):
        self.gradient_instances = ordered_instances
        self.custom_order = custom_order
        self.result_store = ResultStore()  # Use a dedicated result store

    def compute_gradients(self, **kwargs) -> List[Dict]:
        """
        Compute hyper-gradients sequentially using the ordered instances.

        This method processes the hyper-gradients in the defined order, passing intermediate
        results between consecutive gradient operators.

        Parameters
        ----------
        **kwargs : dict
            Additional arguments required for gradient computations.

        Returns
        -------
        List[Dict]
            A list of dictionaries containing results for each gradient operator.
        """
        self.result_store.clear()  # Reset the result store
        intermediate_result = None

        for idx, gradient_instance in enumerate(self.gradient_instances):
            # Compute the gradient, passing the intermediate result as input
            result = gradient_instance.compute_gradients(
                **(kwargs if idx == 0 else intermediate_result),
                next_operation=(
                    self.custom_order[idx + 1]
                    if idx + 1 < len(self.custom_order)
                    else None
                ),
            )
            # Store the result
            self.result_store.add(f"gradient_operator_results_{idx}", result)
            intermediate_result = result

        return self.result_store.get_results()


def makes_functional_hyper_operation(custom_order: List[str], **kwargs) -> SequentialHG:
    """
    Dynamically create a SequentialHG object with ordered gradient operators.

    This function validates the user-defined operator order, adjusts it to conform
    with predefined gradient rules, and dynamically loads the corresponding operator classes.

    Parameters
    ----------
    custom_order : List[str]
        The user-defined order of gradient operators.

    **kwargs : dict
        Additional arguments required for initializing gradient operator instances.

    Returns
    -------
    SequentialHG
        An instance of SequentialHG containing the ordered gradient operators and result management.
    """
    # Load the predefined gradient order
    gradient_order = HyperGradientRules.get_gradient_order()

    # Adjust custom order based on predefined gradient order
    adjusted_order = validate_and_adjust_order(custom_order, gradient_order)

    # Dynamically load classes
    gradient_classes = {}
    # module = importlib.import_module("boat.hyper_ol")
    for op in custom_order:
        gradient_classes[op] = get_registered_operation(op)

    # Reorder classes according to adjusted order
    ordered_instances = [gradient_classes[op](**kwargs) for op in adjusted_order]

    # Return the enhanced sequential hyper-gradient class
    return SequentialHG(ordered_instances, custom_order)


def validate_and_adjust_order(
    custom_order: List[str], gradient_order: List[List[str]]
) -> List[str]:
    """
    Validate and adjust the custom order to match the predefined gradient order.

    This function ensures that the user-defined order adheres to the predefined grouping
    rules and adjusts it accordingly.

    Parameters
    ----------
    custom_order : List[str]
        The user-provided order of gradient operators.

    gradient_order : List[List[str]]
        The predefined order of gradient operator groups.

    Returns
    -------
    List[str]
        Adjusted order of gradient operators following the predefined rules.
    """
    # Create a set of valid operators for quick lookup
    valid_operators = {op for group in gradient_order for op in group}

    # Filter out invalid operators
    custom_order = [op for op in custom_order if op in valid_operators]

    # Adjust order to follow gradient_order
    adjusted_order = []
    for group in gradient_order:
        for op in group:
            if op in custom_order:
                adjusted_order.append(op)

    return adjusted_order
